from zoneinfo import ZoneInfo

import pandas as pd
import pytest

from cosimtlk.simulation.storage import ScheduleStore
from tests.conftest import fake_schedule


@pytest.fixture(scope="function")
def db():
    return ScheduleStore()


def test_store_history_without_tz(db):
    history = fake_schedule(
        made_at=[pd.Timestamp("2021-01-01 00:00:00"), pd.Timestamp("2021-01-01 01:00:00")],
        columns=["a", "b"],
    )
    db.store_history(schedule=history)

    schedule = db.get_schedule("schedule")
    assert isinstance(schedule, pd.DataFrame)
    assert schedule.timestamp.dt.tz == ZoneInfo("UTC")
    assert schedule.made_at.dt.tz == ZoneInfo("UTC")


def test_store_history_with_tz(db):
    history = fake_schedule(
        made_at=[pd.Timestamp("2021-01-01 00:00:00"), pd.Timestamp("2021-01-01 01:00:00")],
        columns=["a", "b"],
        tz=ZoneInfo("Europe/Brussels"),
    )
    db.store_history(schedule=history)

    schedule = db.get_schedule("schedule")
    assert isinstance(schedule, pd.DataFrame)
    assert schedule.timestamp.dt.tz == ZoneInfo("Europe/Brussels")
    assert schedule.made_at.dt.tz == ZoneInfo("Europe/Brussels")


def test_get_set_schedule_without_tz(db):
    made_at = pd.Timestamp("2021-01-01 00:00:00")
    schedule = fake_schedule(made_at)

    db.store_schedule("schedule", schedule, made_at=made_at)

    schedule = db.get_schedule("schedule")
    assert isinstance(schedule, pd.DataFrame)
    assert schedule.timestamp.dt.tz == ZoneInfo("UTC")
    assert schedule.made_at.dt.tz == ZoneInfo("UTC")


def test_get_set_schedule_with_tz(db):
    made_at = pd.Timestamp("2021-01-01 00:00:00", tz=ZoneInfo("Europe/Brussels"))
    schedule = fake_schedule(made_at, tz=ZoneInfo("Europe/Brussels"))

    db.store_schedule("schedule", schedule, made_at=made_at)

    schedule = db.get_schedule("schedule")
    assert isinstance(schedule, pd.DataFrame)
    assert schedule.timestamp.dt.tz == ZoneInfo("Europe/Brussels")
    assert schedule.made_at.dt.tz == ZoneInfo("Europe/Brussels")



from .base import Entity
from .fmu import FMUEntity
from .generic import GenericProcess
from .input import Input
from .sensor import Sensor

__all__ = [
    "Entity",
    "FMUEntity",
    "Input",
    "GenericProcess",
    "Sensor",
]



from datetime import datetime

from pydantic import BaseModel

from cosimtlk.models import FMUInputType


class SimulatorModel(BaseModel):
    id: str
    fmu: str
    created_at: datetime


class SimulatorCreateModel(BaseModel):
    start_values: dict[str, FMUInputType] = {}
    start_time: int = 0
    step_size: int = 1



from typing import Iterable, Optional, Union
from zoneinfo import ZoneInfo

import numpy as np
import pandas as pd
import pytest

from cosimtlk.wrappers import FMIWrapper


@pytest.fixture(scope="function")
def wrapper():
    path = "./fmus/ModSim.Layouts.test.fmu"
    return FMIWrapper(path)


def fake_data(
    start: pd.Timestamp = pd.Timestamp("2020-01-01"),
    freq: str = "1H",
    periods: int = 10,
    tz: Optional[ZoneInfo] = None,
    columns: Iterable[str] = ("a",),
):
    index = pd.date_range(start, freq=freq, periods=periods, tz=tz)
    data = {column: np.random.random(periods) for column in columns}
    df = pd.DataFrame(index=index, data=data)
    return df


def fake_schedule(
    made_at: Union[pd.Timestamp, Iterable[pd.Timestamp]],
    freq: str = "1H",
    periods: int = 10,
    tz: Optional[ZoneInfo] = None,
    columns: Iterable[str] = ("a",),
):
    if isinstance(made_at, pd.Timestamp):
        start = made_at + pd.Timedelta(freq)
        schedule = fake_data(start=start, freq=freq, periods=periods, tz=tz, columns=columns)
        return schedule

    schedules = []
    for made_at_ in made_at:
        start = made_at_ + pd.Timedelta(freq)
        schedule_ = (
            fake_data(start=start, freq=freq, periods=periods, tz=tz, columns=columns)
            .rename_axis("timestamp")
            .reset_index()
        )
        schedule_["made_at"] = made_at_.replace(tzinfo=tz)
        schedules.append(schedule_)

    schedule = pd.concat(schedules, ignore_index=True)
    return schedule



import functools
from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

import cron_converter


def with_tz(dt: datetime, tzinfo: ZoneInfo = ZoneInfo("UTC")) -> datetime:
    """Add timezone information to a datetime object if missing.

    Args:
        dt: Datetime object with or without timezone information.
        tzinfo: Timezone information to add in case the datetime object has no timezone info.
    """
    tzinfo = dt.tzinfo or tzinfo
    return dt.replace(tzinfo=tzinfo)


def every(*, seconds: int = 0, minutes: int = 0, hours: int = 0, days: int = 0):
    """Decorator to schedule a process to run forever with a given delay."""
    total_delay_seconds = seconds + minutes * 60 + hours * 3600 + days * 86400
    assert total_delay_seconds > 0, "At least one time unit must be specified."

    def decorator(func):
        """Decorator to schedule a process to run forever with a given delay."""

        @functools.wraps(func)
        def wrapped(self, *args, **kwargs):
            """Wrapper to schedule a process to run forever with a given delay."""
            while True:
                func(self, *args, **kwargs)
                yield self.env.timeout(total_delay_seconds)

        return wrapped

    return decorator


def cron(*, minute="*", hour="*", day="*", month="*", weekday="*"):
    """Decorator to schedule a process to run forever with a given cron expression."""
    cron_str = " ".join(
        [minute.strip(), hour.strip(), day.strip(), month.strip(), weekday.strip()]
    )
    cron_instance = cron_converter.Cron(cron_str)

    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            # To return the simulation time as first date if exact match
            last_run = self.env.simulation_datetime
            schedule = cron_instance.schedule(last_run - timedelta(seconds=1))
            while True:
                next_run = schedule.next()
                next_in_seconds = int((next_run - last_run).total_seconds())
                yield self.env.timeout(next_in_seconds)
                func(self, *args, **kwargs)
                last_run = next_run

        return wrapper

    return decorator



from typing import Any, Optional
from zoneinfo import ZoneInfo

import pandas as pd
from pandas import DataFrame, Series

from cosimtlk.models import DateTimeLike, FMUInputType


class StateStore:
    def __init__(self, namespace_separator: str = ":") -> None:
        """A simple state store that allows to store and retrieve values by key and a namespace."""
        if namespace_separator is None or namespace_separator == "":
            raise ValueError("The namespace separator cannot be None or an empty string.")
        assert len(namespace_separator) == 1, "The namespace separator must be a single character."
        self.namespace_separator = namespace_separator
        self._state: dict[str, Any] = {}

    def make_namespace(self, *args: str) -> str:
        return self.namespace_separator.join(args)

    def _slip_namespace(self, item: str) -> list[str]:
        return item.split(self.namespace_separator)

    def __getitem__(self, item: str) -> Any:
        if self.namespace_separator in item:
            current_dict = self._state
            keys = self._slip_namespace(item)
            for key in keys[:-1]:
                current_dict = current_dict.get(key, {})
            return current_dict.get(keys[-1])
        return self._state[item]

    def __setitem__(self, key: str, value: Any) -> None:
        if self.namespace_separator in key:
            current_dict = self._state
            keys = self._slip_namespace(key)
            for key in keys[:-1]:
                current_dict = current_dict.setdefault(key, {})
            current_dict[keys[-1]] = value
        else:
            self._state[key] = value

    def __delitem__(self, key: str) -> None:
        if self.namespace_separator in key:
            current_dict = self._state
            keys = self._slip_namespace(key)
            for key in keys[:-1]:
                current_dict = current_dict.get(key, {})
            current_dict.pop(keys[-1], None)
        else:
            self._state.pop(key, None)

    def get_all(self, *, namespace: Optional[str] = None) -> dict[str, Any]:
        if namespace is None:
            return self._state
        current_dict = self._state
        keys = self._slip_namespace(namespace)
        for key in keys:
            current_dict = current_dict.get(key, {})
        return current_dict

    def get(self, key: str, namespace: Optional[str] = None) -> Any:
        if namespace is not None:
            key = self.make_namespace(namespace, key)
        return self.__getitem__(key)

    def delete(self, *key: str, namespace: Optional[str] = None) -> None:
        if namespace is not None:
            key = [f"{namespace}{self.namespace_separator}{k}" for k in key]
        for k in key:
            del self[k]

    def set(self, namespace: Optional[str] = None, **states: Any) -> None:
        for key, value in states.items():
            if namespace is not None:
                key = f"{namespace}{self.namespace_separator}{key}"
            self[key] = value


class ObservationStore:
    def __init__(self) -> None:
        self._db: dict[str, Series] = {}

    @classmethod
    def with_history(cls, **history: Series) -> "ObservationStore":
        store = cls()
        store.store_history(**history)
        return store

    def store_history(self, **history: Series):
        for key, history_ in history.items():
            assert isinstance(history_.index, pd.DatetimeIndex)

            history_.rename_axis("timestamp", inplace=True)
            if history_.index.tz is None:
                history_.index = history_.index.tz_localize(ZoneInfo("UTC"))

            self._db[key] = history_

    def store_observation(self, key: str, value: FMUInputType, ts: DateTimeLike) -> None:
        if ts.tzinfo is None:
            ts = ts.replace(tzinfo=ZoneInfo("UTC"))
        if key not in self._db:
            self._db[key] = pd.Series(name=key, index=[ts], data=[value]).rename_axis("timestamp")
        self._db[key][ts] = value

    def store_observations(self, ts: DateTimeLike, **values: FMUInputType) -> None:
        for key, value in values.items():
            self.store_observation(key, value, ts)

    def get_observations(
        self,
        key: str,
        start: Optional[DateTimeLike] = None,
        end: Optional[DateTimeLike] = None,
        limit: Optional[int] = None,
    ) -> Series:
        obs = self._db[key].loc[start:end]
        if limit is not None:
            obs = obs.tail(limit)
        return obs


class ScheduleStore:
    def __init__(self) -> None:
        self._db: dict[str, DataFrame] = {}

    def store_history(self, **history: DataFrame) -> None:
        for key, schedule in history.items():
            if schedule.timestamp.dt.tz is None:
                schedule.timestamp = schedule.timestamp.dt.tz_localize(ZoneInfo("UTC"))

            if schedule.made_at.dt.tz is None:
                schedule.made_at = schedule.made_at.dt.tz_localize(ZoneInfo("UTC"))

            self._db[key] = schedule

    @classmethod
    def with_history(cls, **history: DataFrame) -> "ScheduleStore":
        store = cls()
        store.store_history(**history)
        return store

    @staticmethod
    def _create_schedule_df(schedule: DataFrame, made_at: DateTimeLike) -> DataFrame:
        if made_at.tzinfo is None:
            made_at = made_at.replace(tzinfo=ZoneInfo("UTC"))

        schedule = schedule.copy().rename_axis("timestamp").reset_index().assign(made_at=made_at)
        if schedule.timestamp.dt.tz is None:
            schedule.timestamp = schedule.timestamp.dt.tz_localize(ZoneInfo("UTC"))
        return schedule

    def store_schedule(self, key: str, schedule: DataFrame, made_at: DateTimeLike) -> None:
        schedule = self._create_schedule_df(schedule, made_at)

        if key not in self._db:
            self._db[key] = schedule

        self._db[key] = pd.concat([self._db[key], schedule])

    def store_schedules(self, made_at: DateTimeLike, **schedules: DataFrame) -> None:
        for key, schedule in schedules.items():
            self.store_schedule(key, schedule, made_at)

    def get_schedule(
        self,
        key: str,
        made_after: Optional[DateTimeLike] = None,
        made_before: Optional[DateTimeLike] = None,
    ) -> DataFrame:
        if made_after is None and made_before is None:
            return self._db[key]

        if made_after is None and made_before is not None:
            mask = self._db[key].index <= made_before
        elif made_after is not None and made_before is None:
            mask = self._db[key].index >= made_after
        else:
            mask = (self._db[key].index >= made_after) & (self._db[key].index <= made_before)
        return self._db[key].loc[mask]

    def get_last_schedule(self, key: str, made_at: DateTimeLike) -> FMUInputType:
        last_made_at = (self._db[key]["made_at"].unique() <= made_at).max()
        return self._db[key].loc["made_at" == last_made_at, key]



from datetime import datetime, timedelta
from typing import Optional, Union
from zoneinfo import ZoneInfo

import numpy as np
from simpy.util import start_delayed

from cosimtlk.models import DateTimeLike
from cosimtlk.simulation.entities import Entity
from cosimtlk.simulation.environment import Environment
from cosimtlk.simulation.utils import with_tz


class SimulationRunner:
    def __init__(
        self,
        *,
        initial_time: DateTimeLike,
        entities: Optional[list[Entity]] = None,
    ) -> None:
        """Simulation runner.

        Organizes the entities processes into a discrete event simulation. If multiple entities would run at the same
        time, the order is determined by the order of the entities in the entities list.

        Args:
            initial_time: The initial time of the simulation. Can be either a datetime or a Timestamp.
            entities: A list of entities inside the simulation.
        """
        initial_time = self._parse_datetime(initial_time)
        initial_timestamp = self._dt_to_timestamp(initial_time)
        self.tzinfo: ZoneInfo = initial_time.tzinfo

        # Create simulation environment
        self._environment = Environment(initial_time=initial_timestamp, tzinfo=self.tzinfo)

        # Initialize processes
        processes = []
        for entity in entities or []:
            processes.extend(entity.initialize(self._environment))
        process_delays = np.linspace(0.005, 0.995, len(processes))

        for process, delay in zip(processes, process_delays, strict=True):
            start_delayed(self._environment, process(), delay)

    def __repr__(self) -> str:
        return f"<SimulationRunner t={self._environment.simulation_timestamp}>"

    @staticmethod
    def _parse_datetime(dt: DateTimeLike) -> DateTimeLike:
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=ZoneInfo("UTC"))
        return dt

    @staticmethod
    def _dt_to_timestamp(dt: DateTimeLike) -> int:
        """Converts a datetime to a unix timestamp."""
        return int(dt.timestamp())

    @staticmethod
    def _td_to_duration(td: timedelta) -> int:
        """Converts a timedelta to a duration in seconds."""
        return int(td.total_seconds())

    def run(self, until: Union[int, datetime], progress_bar: bool = True):
        """Runs the simulation until the given timestamp.

        Args:
            until: The timestamp until the simulation should run.
            progress_bar: Whether to show a progress bar.
        """
        if isinstance(until, datetime):
            if until.tzinfo is not None and until.tzinfo != self.tzinfo:
                raise ValueError(
                    f"Until must be in the same timezone as the initial time. {self.tzinfo} != {until.tzinfo}"
                )
            until = with_tz(until, tzinfo=self.tzinfo)
            until = self._dt_to_timestamp(until)
        else:
            until = int(until)
        self._environment.run(until=until, progress_bar=progress_bar)

    def run_for(self, duration: Union[int, timedelta], progress_bar: bool = True):
        """Runs the simulation for the given duration.

        Args:
            duration: The duration for which the simulation should run.
            progress_bar: Whether to show a progress bar.
        """
        if isinstance(duration, timedelta):
            duration = self._td_to_duration(duration)
        else:
            duration = int(duration)
        self.run(
            until=self._environment.simulation_timestamp + duration, progress_bar=progress_bar
        )



from __future__ import annotations

from abc import ABCMeta
from queue import Empty, Queue
from typing import Protocol

from cosimtlk.simulation.entities import Entity


class Subscriber(Protocol):
    def notify(self, event: dict):
        ...


class EventBus:
    def __init__(self):
        self._subscribers: dict[str, list] = {}

    def subscribe(self, event_type: str, subscriber: Subscriber):
        if event_type not in self._subscribers:
            self._subscribers[event_type] = []
        self._subscribers[event_type].append(subscriber)

    def publish(self, event_type: str, message: dict) -> None:
        if event_type not in self._subscribers:
            return

        for subscriber in self._subscribers[event_type]:
            subscriber.notify(message)


class Observer(Entity, metaclass=ABCMeta):
    def __init__(self, name: str):
        super().__init__(name)
        self.queue: Queue = Queue()

    def notify(self, message: dict):
        return self.queue.put(message)

    def consume(self) -> dict | None:
        try:
            message = self.queue.get(block=False)
        except Empty:
            message = None
        return message



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Callable, Generator, Optional

if TYPE_CHECKING:
    from cosimtlk.simulation import Environment


class Entity(metaclass=ABCMeta):
    def __init__(self, name: str) -> None:
        """Entity base class that defines the interface for all entities.

        Abstract base class for all entities in the simulation. An entity is
        a component that can be scheduled in the simulation environment.
        It must implement the `processes` method that returns a list of
        processes that should be scheduled.
        The order of the processes in the list is the order in which they
        will be scheduled in case they should fire exactly the same time.

        Args:
            name: Name of the entity for identification purposes.
        """
        self._name = name
        self._env: Optional[Environment] = None

    def __repr__(self):
        """Representation of the entity."""
        return f"<{self.__class__.__name__}: {self.name}>"

    @property
    def name(self) -> str:
        """Name of the entity."""
        return self._name

    @property
    def env(self) -> Environment:
        """Simulation environment."""
        return self._env

    def initialize(self, env: Environment) -> list[Callable[[], Generator]]:
        """Initialize the entity in the simulation.

        Returns:
            List of processes that should be scheduled.
        """
        self._env = env
        return self.processes

    @property
    @abstractmethod
    def processes(self) -> list[Callable[[], Generator]]:
        """List of processes that should be scheduled.

        Returns:
            List of processes.
        """
        raise NotImplementedError



from __future__ import annotations

import functools
from typing import Callable, Generator

from cosimtlk.simulation.entities import Entity
from cosimtlk.simulation.storage import ObservationStore, StateStore


class Sensor(Entity):
    def __init__(
        self,
        name: str,
        *,
        measurements: dict[str, str],
        scheduler: Callable,
        state_store: StateStore,
        db: ObservationStore,
    ):
        """An entity that stores the current state of the simulation into long term storage.

        Args:
            name: The name of the entity.
            measurements: A mapping of state names to measurement names. The keys are used as
                the names of the measurements inside the database, while the values determine the state.
            scheduler: A generator function that schedules a function such as `cosimtlk.simulation.utils.every`
                or `cosimtlk.simulation.utils.cron`.
            state_store: The state store to get the state from.
            db: The database to store the observations in.
        """
        super().__init__(name)
        self.measurements = measurements
        self.scheduler = scheduler
        self.state_store = state_store
        self.db = db

    @property
    def processes(self) -> list[Callable[[], Generator]]:
        scheduled_process = self.scheduler(self.__class__.sensing_process)
        return [
            functools.partial(scheduled_process, self),
        ]

    def sensing_process(self):
        values = {
            measurement_name: self.state_store.get(state_name)
            for measurement_name, state_name in self.measurements.items()
        }
        timestamp = self.env.simulation_datetime
        self.db.store_observations(timestamp, **values)



from datetime import datetime
from typing import Any, Optional, Union
from zoneinfo import ZoneInfo

import simpy as sp
from simpy import Event
from simpy.core import EmptySchedule, SimTime, StopSimulation
from simpy.events import URGENT
from tqdm import tqdm


class Environment(sp.Environment):
    def __init__(self, initial_time: int = 0, tzinfo: ZoneInfo = ZoneInfo("UTC")):
        """Execution environment for an event-based simulation. The passing of time
        is simulated by stepping from event to event.

        This subclass of simpy.Environment and makes the following changes:
        - The simulation time is stored as a unix timestamp and is fixed at second resolution.
          The sub-second resolution is used to set the priority of events which would fire at the same time.
        - The simulation is also aware of the timezone, and the current time can be accessed
          as a timezone aware datetime object.

        Args:
            initial_time: The initial time of the simulation as a unix timestamp.
            tzinfo: The timezone of the simulation.
        """
        super().__init__(initial_time=initial_time)
        self.tzinfo = tzinfo

    @property
    def now(self) -> SimTime:
        """The current simulation time as a float.

        This property is not used in this subclass and is made internal. Use simulation_timestamp instead.
        """
        raise AttributeError(
            "Now is used internally, use simulation_time instead to access the current time."
        )

    @property
    def simulation_timestamp(self) -> int:
        """The current simulation time as a unix timestamp."""
        return int(self._now)

    @property
    def simulation_datetime(self) -> datetime:
        """The current simulation time as a timezone aware datetime object."""
        return datetime.fromtimestamp(int(self._now), tz=self.tzinfo)

    def run(
        self,
        until: Union[SimTime, Event, None] = None,
        progress_bar: bool = True,
    ) -> Optional[Any]:
        """Run the environment until the given event or time.

        Args:
            until: The event or time until which the environment should be run.
            progress_bar: Whether to show a progress bar.

        Returns:
            The value of the event if it was triggered, otherwise None.
        """
        if until is not None:
            if not isinstance(until, Event):
                # Assume that *until* is a number if it is not None and
                # not an event.  Create a Timeout(until) in this case.
                at: SimTime
                if isinstance(until, int):
                    at = until
                else:
                    at = float(until)

                if at <= self._now:
                    raise ValueError(f"until(={at}) must be > the current simulation time.")

                # Schedule the event before all regular timeouts.
                until = Event(self)
                until._ok = True
                until._value = None
                self.schedule(until, URGENT, at - self._now)

            elif until.callbacks is None:
                # Until event has already been processed.
                return until.value

            until.callbacks.append(StopSimulation.callback)

        try:
            if progress_bar:
                pbar = tqdm(total=at - self._now, desc="Simulation progress", unit="s")
                while True:
                    now = self._now
                    self.step()
                    progress = self._now - now
                    pbar.update(progress)
            else:
                while True:
                    self.step()
        except StopSimulation as exc:
            return exc.args[0]  # == until.value
        except EmptySchedule as e:
            if until is not None:
                assert not until.triggered
                raise RuntimeError(
                    f'No scheduled events left but "until" event was not ' f"triggered: {until}"
                ) from e
        return None



import logging
from typing import Any, Optional, Union

from cosimtlk.app.client import SimulatorClient
from cosimtlk.models import FMUInputType
from cosimtlk.wrappers.base import Wrapper

logger = logging.getLogger(__name__)


class RemoteFMIWrapper(Wrapper):
    def __init__(self, path: str, client: SimulatorClient):
        self._fmu_path = path
        self.client = client
        self._id = None
        self._step_size = 0
        self._current_time = 0

    def initialize(
        self,
        start_values: Optional[dict[str, FMUInputType]] = None,
        start_time: Union[int, float] = 0,
        step_size: Union[int, float] = 1,
    ) -> None:
        simulator = self.client.create_simulator(
            self._fmu_path, start_values=start_values, start_time=start_time, step_size=step_size
        )
        self._id = simulator.id
        self._step_size = step_size
        self._current_time = start_time

    def __repr__(self):
        return f"{self.__class__.__name__}(path={self._fmu_path}>"

    def __del__(self):
        self.close()

    @property
    def step_size(self) -> int:
        return self._step_size

    @property
    def current_time(self) -> int:
        return self._current_time

    @property
    def is_initialized(self) -> bool:
        return self._id is not None

    def info(self) -> dict[str, Any]:
        self._check_is_initialized()
        return self.client.get_fmu_info(self._fmu_path)

    def close(self):
        if self.is_initialized:
            self.client.delete_simulator(self._id)
            self._id = None

    def reset(
        self,
        *,
        start_values: Optional[dict[str, FMUInputType]] = None,
        start_time: Union[int, float] = 0,
        step_size: Union[int, float] = 1,
    ) -> None:
        self._check_is_initialized()
        self.client.reset(
            self._id, start_values=start_values, start_time=start_time, step_size=step_size
        )
        self._step_size = step_size
        self._current_time = start_time

    def _check_is_initialized(self):
        if self._id is None:
            raise RuntimeError("FMU is not initialized.")

    def step(
        self, *, input_values: Optional[dict[str, FMUInputType]] = None
    ) -> dict[str, FMUInputType]:
        self._check_is_initialized()
        result = self.client.step(self._id, input_values=input_values)
        self._current_time = result["current_time"]
        return result

    def advance(
        self, until: int, *, input_values: Optional[dict[str, FMUInputType]] = None
    ) -> dict[str, FMUInputType]:
        self._check_is_initialized()
        result = self.client.advance(self._id, until, input_values=input_values)
        self._current_time = result["current_time"]
        return result

    def read_outputs(self) -> dict[str, FMUInputType]:
        self._check_is_initialized()
        return self.client.get_outputs(self._id)

    def change_parameters(self, parameters: dict[str, FMUInputType]) -> None:
        self._check_is_initialized()
        self.client.change_parameters(self._id, parameters=parameters)



import logging
from pathlib import Path
from typing import Annotated

from fastapi import APIRouter, Depends
from starlette.responses import Response

from cosimtlk.app.dependencies import get_fmu_dir
from cosimtlk.wrappers.local import FMIWrapper

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/fmus", tags=["FMUs"])


@router.get("/", description="List available FMUs")
def list(fmu_dir: Annotated[Path, Depends(get_fmu_dir)]):
    return {
        "path": fmu_dir,
        "fmus": sorted([fmu.stem for fmu in fmu_dir.glob("*.fmu")]),
    }


@router.get("/{fmu}/info", description="Get information about an FMU")
def get_info(fmu: str, fmu_dir: Annotated[Path, Depends(get_fmu_dir)]):
    fmu_path = fmu_dir / f"{fmu}.fmu"
    if not fmu_path.exists():
        return Response(status_code=404)
    return FMIWrapper(fmu_path).info()



from abc import ABCMeta, abstractmethod
from typing import Any, Optional, Union

from cosimtlk.models import FMUInputType


class Wrapper(metaclass=ABCMeta):
    def __enter__(self) -> "Wrapper":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    @abstractmethod
    def info(self) -> dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    def initialize(
        self,
        *,
        start_values: Optional[dict[str, FMUInputType]] = None,
        start_time: Union[int, float] = 0,
        step_size: Union[int, float] = 1,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def reset(
        self,
        *,
        start_values: Optional[dict[str, FMUInputType]] = None,
        start_time: Union[int, float] = 0,
        step_size: Union[int, float] = 1,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def step(
        self,
        *,
        input_values: Optional[dict[str, FMUInputType]] = None,
    ) -> dict[str, FMUInputType]:
        raise NotImplementedError

    @abstractmethod
    def advance(
        self,
        until: int,
        *,
        input_values: Optional[dict[str, FMUInputType]] = None,
    ) -> dict[str, FMUInputType]:
        raise NotImplementedError

    @abstractmethod
    def change_parameters(
        self,
        parameters: dict[str, FMUInputType],
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def read_outputs(self) -> dict[str, FMUInputType]:
        raise NotImplementedError



import logging
from pathlib import Path

from fastapi import APIRouter, Body, Response
from starlette.responses import JSONResponse

from cosimtlk.app.config import settings
from cosimtlk.app.schemas import SimulatorCreateModel, SimulatorModel
from cosimtlk.app.services.simulator import simulator_service
from cosimtlk.models import FMUInputType

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/simulators", tags=["Simulators"])


@router.get("/", response_model=list[SimulatorModel])
def list_simulators():
    return [SimulatorModel(**simulator) for simulator in simulator_service.list()]


@router.post("/", response_model=SimulatorModel)
def create_simulator(
    fmu: str,
    data: SimulatorCreateModel,
):
    fmu_path = Path(settings.fmu_dir).resolve() / f"{fmu}.fmu"
    if not fmu_path.exists():
        return Response(status_code=404)

    simulator = simulator_service.create(
        path=fmu_path,
        start_values=data.start_values,
        start_time=data.start_time,
        step_size=data.step_size,
    )
    return simulator


@router.get("/{id}", response_model=SimulatorModel)
def get_simulator(id: str):
    try:
        return simulator_service.get(id)
    except KeyError:
        return Response(status_code=404)


@router.delete("/{id}")
def delete_simulator(id: str):
    try:
        simulator_service.delete(id)
        return Response(status_code=204)
    except KeyError:
        return Response(status_code=404)


@router.get("/{id}/outputs")
def read_outputs(id: str):
    try:
        simulator = simulator_service.get_simulator(id)
        outputs = simulator.read_outputs()
    except KeyError:
        return Response(status_code=404)
    return JSONResponse(status_code=200, content=outputs)


@router.post("/{id}/step")
def step(
    id: str,
    input_values: dict[str, FMUInputType] = Body({}),
):
    try:
        simulator = simulator_service.get_simulator(id)
        result = simulator.step(input_values=input_values)
    except Exception as e:
        logger.exception(e)
        return JSONResponse(status_code=500, content={"error": str(e)})
    return JSONResponse(status_code=200, content=result)


@router.post("/{id}/advance")
def advance(id: str, until: int, input_values: dict[str, FMUInputType] = Body({})):
    try:
        simulator = simulator_service.get_simulator(id)
        result = simulator.advance(until, input_values=input_values)
    except ValueError as e:
        logger.exception(e)
        return JSONResponse(status_code=500, content={"error": str(e)})
    return JSONResponse(status_code=200, content=result)


@router.put("/{id}/parameters")
def change_parameters(id: str, parameters: dict[str, FMUInputType] = Body({})):
    try:
        simulator = simulator_service.get_simulator(id)
        simulator.change_parameters(parameters)
    except ValueError as e:
        logger.exception(e)
        return JSONResponse(status_code=500, content={"error": str(e)})
    return JSONResponse(status_code=200, content={"message": "Success"})


@router.post("/{id}/reset")
def reset(
    id: str,
    data: SimulatorCreateModel,
):
    try:
        simulator = simulator_service.get_simulator(id)
        simulator.reset(
            start_values=data.start_values, start_time=data.start_time, step_size=data.step_size
        )
    except Exception as e:
        logger.exception(e)
        return JSONResponse(status_code=500, content={"error": str(e)})
    return JSONResponse(status_code=200, content={"message": "Success"})



import functools
from typing import Callable, Generator

from cosimtlk.simulation.entities import Entity


class GenericProcess(Entity):
    def __init__(
        self,
        name: str,
        func: Callable,
        *,
        scheduler: Callable,
        **kwargs,
    ):
        """A generic entity that can schedule the given function.

        Args:
            name: Name of the controller for identification purposes.
            func: The function to be scheduled.
            scheduler: A scheduler function that returns a generator such as 'every' or 'cron' from utils.
        """
        super().__init__(name)
        self.scheduler = scheduler
        self._func = func
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def processes(self) -> list[Callable[[], Generator]]:
        decorated = self.scheduler(self._func)
        return [
            functools.partial(decorated, self),
        ]



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Callable, Generator

from cosimtlk.models import FMUInputType
from cosimtlk.simulation.entities import Entity
from cosimtlk.wrappers import FMIWrapper

if TYPE_CHECKING:
    from cosimtlk.simulation.storage import StateStore

logger = logging.getLogger(__name__)


class FMUEntity(Entity):
    def __init__(
        self,
        name: str,
        *,
        simulator: FMIWrapper,
        start_values: dict[str, FMUInputType],
        step_size: int,
        simulation_step_size: int,
        state_store: StateStore,
        namespace: str | None = None,
        input_namespace: str = "inputs",
        output_namespace: str = "outputs",
    ):
        super().__init__(name)
        # Simulator inputs
        self.simulator = simulator
        self.start_values = start_values
        self.step_size = step_size
        self.simulation_step_size = simulation_step_size

        self.state_store = state_store
        self.namespace = namespace or self.name
        self.input_namespace = self.state_store.make_namespace(self.namespace, input_namespace)
        self.output_namespace = self.state_store.make_namespace(self.namespace, output_namespace)

    @property
    def processes(self) -> list[Callable[[], Generator]]:
        return [self.simulation_process]

    def _store_outputs(self, outputs):
        self.state_store.set(**outputs, namespace=self.output_namespace)
        logger.debug(f"{self}: t={self.env.simulation_datetime}, outputs={outputs}")

    def simulation_process(self):
        self.simulator.initialize(
            start_values=self.start_values,
            step_size=self.step_size,
            start_time=self.env.simulation_timestamp,
        )
        self._store_outputs(self.simulator.read_outputs())
        while True:
            simulation_timestamp = self.env.simulation_timestamp
            # Collect inputs
            inputs = self.state_store.get_all(namespace=self.input_namespace)
            logger.debug(f"{self}: t={self.env.simulation_datetime}, inputs={inputs}")
            # Advance simulation
            outputs = self.simulator.advance(
                simulation_timestamp + self.simulation_step_size, input_values=inputs
            )
            # Wait until next step
            time_until_next_step = self.simulator.current_time - simulation_timestamp
            yield self.env.timeout(time_until_next_step)
            # Store outputs
            self._store_outputs(outputs)



from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    fmu_dir: str = "./fmus"

    class Config:
        env_prefix = "COSIMTLK_"  # defaults to no prefix, i.e. ""


settings = Settings()



import pytest


def test_initialize(wrapper):
    assert wrapper.is_initialized is False
    assert wrapper.step_size == 1
    assert wrapper.current_time == 0

    wrapper.initialize()
    assert wrapper.is_initialized is True
    assert wrapper.step_size == 1
    assert wrapper.current_time == 0

    wrapper.close()
    assert wrapper.is_initialized is False
    assert wrapper.step_size == 1
    assert wrapper.current_time == 0


def test_step_on_closed_raises(wrapper):
    with pytest.raises(RuntimeError):
        """Cannot call step() on a uninitialized fmu."""
        wrapper.step()


def test_advance_on_closed_raises(wrapper):
    with pytest.raises(RuntimeError):
        """Cannot call advance() on a uninitialized fmu."""
        wrapper.advance(3)


def test_reset_on_closed_raises(wrapper):
    with pytest.raises(RuntimeError):
        """Cannot call advance() on a uninitialized fmu."""
        wrapper.reset()


def test_context_manager(wrapper):
    with wrapper() as fmu:
        assert fmu.is_initialized is True
        assert fmu.step_size == 1
        assert fmu.current_time == 0

    assert wrapper.is_initialized is False
    assert wrapper.step_size == 1
    assert wrapper.current_time == 0


def test_step(wrapper):
    with wrapper(
        start_values={
            "integrator.k": 2.0,
            "integrator.y_start": 1.05,
        }
    ) as fmu:
        # First step
        outputs = fmu.step(
            input_values={
                "real_setpoint": 2.0,
                "int_setpoint": 3,
                "bool_setpoint": False,
            }
        )
        expected_output = {
            "current_time": 1,
            "real_output": 5.05,
            "int_output": 3,
            "bool_output": False,
        }
        assert outputs == expected_output

        # Second step
        outputs = fmu.step(
            input_values={
                "real_setpoint": 1.0,
                "int_setpoint": 2,
                "bool_setpoint": True,
            }
        )
        expected_output = {
            "current_time": 2,
            "real_output": 7.05,
            "int_output": 2,
            "bool_output": True,
        }
        assert outputs == expected_output


def test_advance(wrapper):
    with wrapper(
        start_values={
            "integrator.k": 1.0,
            "integrator.y_start": 1.05,
        }
    ) as fmu:
        outputs = fmu.advance(
            10,
            input_values={
                "real_setpoint": 1.0,
                "int_setpoint": 3,
                "bool_setpoint": True,
            },
        )
        expected_output = {
            "current_time": 10,
            "real_output": 11.05,
            "int_output": 3,
            "bool_output": True,
        }
        assert outputs == expected_output


def test_read_outputs(wrapper):
    with wrapper(
        start_values={
            "integrator.k": 2.0,
            "integrator.y_start": 1.05,
            "real_setpoint": 2.0,
            "int_setpoint": 3,
            "bool_setpoint": True,
        }
    ) as fmu:
        outputs = fmu.read_outputs()
        expected_output = {
            "current_time": 0,
            "real_output": 1.05,
            "int_output": 3,
            "bool_output": True,
        }
        assert outputs == expected_output


def test_step_with_custom_stepsize(wrapper):
    with wrapper(
        start_values={
            "integrator.k": 2.0,
            "integrator.y_start": 1.05,
        },
        step_size=5,
        start_time=1,
    ) as fmu:
        # First step
        outputs = fmu.step(
            input_values={
                "real_setpoint": 1.0,
                "int_setpoint": 3,
                "bool_setpoint": True,
            }
        )
        expected_output = {
            "current_time": 6,
            "real_output": 11.05,
            "int_output": 3,
            "bool_output": True,
        }
        assert outputs == expected_output


def test_advance_with_custom_stepsize(wrapper):
    with wrapper(
        start_values={
            "integrator.k": 2.0,
            "integrator.y_start": 1.05,
        },
        step_size=5,
    ) as fmu:
        # First step
        outputs = fmu.advance(
            11,
            input_values={
                "real_setpoint": 1.0,
                "int_setpoint": 3,
                "bool_setpoint": True,
            },
        )
        expected_output = {
            "current_time": 15,
            "real_output": 31.05,
            "int_output": 3,
            "bool_output": True,
        }
        assert outputs == expected_output


def test_change_parameters(wrapper):
    with wrapper(
        start_values={
            "integrator.k": 2.0,
            "integrator.y_start": 1.05,
        }
    ) as fmu:
        outputs = fmu.step(
            input_values={
                "real_setpoint": 1.0,
                "int_setpoint": 3,
                "bool_setpoint": True,
            }
        )
        expected_output = {
            "current_time": 1,
            "real_output": 3.05,
            "int_output": 3,
            "bool_output": True,
        }
        assert outputs == expected_output

        fmu.change_parameters(
            {
                "integrator.k": 1.0,
                # FIXME: this is not ideal that we have to manually carry over state from the previous simulation
                "integrator.y_start": outputs["real_output"],
            }
        )

        outputs = fmu.read_outputs()
        expected_output = {
            "current_time": 1,
            "real_output": 3.05,
            "int_output": 3,
            "bool_output": True,
        }
        assert outputs == expected_output
        print(outputs)

        outputs = fmu.step(
            input_values={
                "real_setpoint": 1.0,
                "int_setpoint": 2,
                "bool_setpoint": False,
            }
        )
        expected_output = {
            "current_time": 2,
            "real_output": 4.05,
            "int_output": 2,
            "bool_output": False,
        }
        assert outputs == expected_output
        print(outputs)


if __name__ == "__main__":
    pytest.main()



from datetime import datetime
from zoneinfo import ZoneInfo

import pandas as pd
import pytest

from cosimtlk.simulation.storage import ObservationStore
from tests.conftest import fake_data


@pytest.fixture(scope="function")
def db():
    return ObservationStore()


def test_store_history_without_tz(db):
    history = fake_data(columns=["a"])

    db.store_history(a=history.a)

    a = db.get_observations("a")
    assert isinstance(a, pd.Series)
    assert isinstance(a.index, pd.DatetimeIndex)
    assert a.index.name == "timestamp"
    assert a.index.tzinfo == ZoneInfo("UTC")


def test_store_history_with_tz(db):
    history = fake_data(columns=["a"], tz=ZoneInfo("Europe/Brussels"))

    db.store_history(a=history.a)

    a = db.get_observations("a")
    assert isinstance(a, pd.Series)
    assert isinstance(a.index, pd.DatetimeIndex)
    assert a.index.name == "timestamp"
    assert a.index.tzinfo == ZoneInfo("Europe/Brussels")


def test_get_set_observation_without_tz(db):
    db.store_observation("a", 1, datetime.fromisoformat("2021-01-01 00:00:00"))
    db.store_observation("a", 2, pd.Timestamp("2021-01-01 01:00:00"))

    observations = db.get_observations("a")
    assert len(observations) == 2
    assert all(observations.values == [1, 2])
    assert observations.name == "a"
    assert observations.index.name == "timestamp"
    assert observations.index.tzinfo == ZoneInfo("UTC")


def test_get_set_observation_with_tz(db):
    db.store_observation("a", 1, datetime.now(tz=ZoneInfo("Europe/Brussels")))
    db.store_observation("a", 2, pd.Timestamp.now(tz=ZoneInfo("Europe/Brussels")))

    observations = db.get_observations("a")
    assert len(observations) == 2
    assert all(observations.values == [1, 2])
    assert observations.name == "a"
    assert observations.index.name == "timestamp"
    assert observations.index.tzinfo == ZoneInfo("Europe/Brussels")



import logging
from typing import Callable, Generator, Union

from pandas import DataFrame, Series

from cosimtlk.simulation.entities import Entity
from cosimtlk.simulation.storage import StateStore

logger = logging.getLogger(__name__)


class Input(Entity):
    def __init__(self, name: str, *, data: Union[Series, DataFrame], state_store: StateStore):
        """An entity that sets the input of the simulator based on the input date.

        Args:
            name: The name of the entity.
            data: The input data. Can be a Series or a DataFrame with a DatetimeIndex.
                The index of the data is used as the time at which the input is set,
                while the name of the columns are used as the name of the inputs.
            state_store: The state store to set the input values in.

        """
        super().__init__(name)
        if isinstance(data, Series):
            data = data.copy().to_frame()
        self.data = data
        self.state_store = state_store
        self._index = 0

    @property
    def processes(self) -> list[Callable[[], Generator]]:
        return [self.set_inputs_process]

    def set_inputs_process(self):
        while True:
            if self.data.empty or self._index >= len(self.data):
                logger.warning(f"{self}: t={self.env.simulation_datetime}, no data left.")
                break

            current_time = self.env.simulation_datetime
            next_point_at = self.data.index[self._index]

            if next_point_at <= current_time:
                next_points = self.data.iloc[self._index].to_dict()
                logger.debug(
                    f"{self}: t={self.env.simulation_datetime}, setting inputs: {next_points}"
                )
                self.state_store.set(**next_points)
                self._index += 1
            else:
                next_point_in = int((next_point_at - current_time).total_seconds())
                yield self.env.timeout(next_point_in)



import logging

from fastapi import FastAPI

from cosimtlk.app.routers import fmus, simulators

logger = logging.getLogger(__name__)

app = FastAPI(title="FMU Simulator")
app.include_router(fmus.router)
app.include_router(simulators.router)


@app.on_event("shutdown")
def shutdown_event():
    from cosimtlk.app.services.simulator import simulator_service

    simulator_service.close()



from datetime import datetime
from enum import Enum
from typing import Union

from pandas import Timestamp

FMUInputType = Union[float, int, str, bool]
DateTimeLike = Union[datetime, Timestamp]


class FMUCausaltyType(str, Enum):
    INPUT = "input"
    OUTPUT = "output"
    PARAMETER = "parameter"
    CALCULATED_PARAMETER = "calculatedParameter"
    LOCAL = "local"



import pytest

from cosimtlk.simulation.storage import StateStore


@pytest.fixture(scope="function")
def state_store():
    return StateStore()


def test_make_namespace(state_store):
    assert state_store.make_namespace("a") == "a"
    assert state_store.make_namespace("a", "b") == "a:b"
    assert state_store.make_namespace("a", "b", "c") == "a:b:c"


def test_setitem_getitem(state_store):
    state_store["a"] = 1
    assert state_store["a"] == 1
    state_store["a"] = 2
    assert state_store["a"] == 2


def test_setitem_getitem_with_namespace(state_store):
    state_store["a:b"] = 1
    assert state_store["a:b"] == 1
    assert state_store["a"] == {"b": 1}

    state_store["a:b"] = 2
    assert state_store["a:b"] == 2
    assert state_store["a"] == {"b": 2}


def test_get_all(state_store):
    state_store["a"] = 1
    state_store["b"] = 2
    assert state_store.get_all() == {"a": 1, "b": 2}


def test_get_all_with_namespace(state_store):
    state_store["a:b"] = 1
    state_store["a:c"] = 2
    assert state_store.get_all(namespace="a") == {"b": 1, "c": 2}


def test_set_get_single(state_store):
    state_store.set(a=1)
    assert state_store.get("a") == 1
    state_store.set(a=2)
    assert state_store.get("a") == 2


def test_set_get_multiple(state_store):
    state_store.set(a=1, b=2)
    assert state_store.get("a") == 1
    assert state_store.get("b") == 2
    state_store.set(a=2, b=3)
    assert state_store.get("a") == 2
    assert state_store.get("b") == 3


def test_set_get_single_with_namespace(state_store):
    state_store.set(namespace="a", b=1)
    assert state_store.get("a:b") == 1
    assert state_store.get("b", namespace="a") == 1


if __name__ == "__main__":
    pytest.main()



import logging
import shutil
from pathlib import Path
from typing import Any, Optional, Union
from uuid import uuid4

import fmpy
from fmpy.fmi2 import FMU2Slave
from fmpy.model_description import ModelDescription, ScalarVariable

from cosimtlk.models import FMUCausaltyType, FMUInputType
from cosimtlk.wrappers.base import Wrapper

logger = logging.getLogger(__name__)


class FMIWrapper(Wrapper):
    def __init__(
        self,
        path: Union[Path, str],
    ):
        """Simulator using an FMU.

        Args:
            path: Path of the FMU file.
        """
        # Path to the FMU file
        fmu_path = Path(path).resolve()
        if not fmu_path.exists():
            raise FileNotFoundError(f"FMU file not found: {fmu_path}")
        if not fmu_path.suffix == ".fmu":
            raise ValueError(f"FMU file must have .fmu extension: {fmu_path}")
        self.fmu_path = str(fmu_path)

        # Read model description from the FMU
        self.model_description: ModelDescription = fmpy.read_model_description(self.fmu_path)

        # Create input and output maps
        self._input_variables: dict[str, ScalarVariable] = {
            variable.name: variable
            for variable in self.model_description.modelVariables
            if variable.causality == FMUCausaltyType.INPUT
        }
        self._output_variables: dict[str, ScalarVariable] = {
            variable.name: variable
            for variable in self.model_description.modelVariables
            if variable.causality == FMUCausaltyType.OUTPUT
        }
        self._model_parameters: dict[str, ScalarVariable] = {
            variable.name: variable
            for variable in self.model_description.modelVariables
            if variable.causality == FMUCausaltyType.PARAMETER
        }

        # Create maps for faster read of outputs
        self._output_names: dict[str, list[str]] = {
            "Real": [],
            "Integer": [],
            "Boolean": [],
            "String": [],
        }
        self._output_refs: dict[str, list[int]] = {
            "Real": [],
            "Integer": [],
            "Boolean": [],
            "String": [],
        }
        for variable in self.model_description.modelVariables:
            if (
                variable.causality == FMUCausaltyType.OUTPUT
                and variable.name in self._output_variables
            ):
                self._output_names[variable.type].append(variable.name)
                self._output_refs[variable.type].append(variable.valueReference)

        self._instantiated = False
        self._initialized = False
        self.fmu: Optional[FMU2Slave] = None
        self._current_time = 0
        self._step_size = 1

    def _instantiate(self) -> None:
        # Check platform compatibility and recompile if necessary
        self._check_platform_compatibility()

        self._unzipdir = fmpy.extract(self.fmu_path)
        self.fmu = FMU2Slave(
            unzipDirectory=self._unzipdir,
            guid=self.model_description.guid,
            instanceName=f"{uuid4()}",
            modelIdentifier=self.model_description.coSimulation.modelIdentifier,
        )

        # Instantiate FMU
        self.fmu.instantiate(visible=False, callbacks=None, loggingOn=False)
        self._instantiated = True

    def __call__(
        self,
        *,
        start_values: Optional[dict[str, FMUInputType]] = None,
        start_time: Union[int, float] = 0,
        step_size: Union[int, float] = 1,
    ):
        self.initialize(start_values=start_values, start_time=start_time, step_size=step_size)
        return self

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(path={self.fmu_path!r})"

    def __del__(self):
        self.close()
        self._free_instance()

    def _check_platform_compatibility(self) -> None:
        if not self._current_platform_is_supported():
            logger.warning(f"{fmpy.platform} is not supported by this FMU, recompiling...")
            fmpy.util.compile_platform_binary(self.fmu_path)
            logger.info(f"Recompiled FMU for platform {fmpy.platform}.")

    def _current_platform_is_supported(self) -> bool:
        current_platform = fmpy.platform
        supported_platforms = fmpy.supported_platforms(self.fmu_path)
        return current_platform in supported_platforms

    @property
    def step_size(self) -> int:
        return self._step_size

    @property
    def current_time(self) -> Union[int, float]:
        return self._current_time

    @property
    def is_initialized(self) -> bool:
        return self._initialized

    def info(self) -> dict[str, Any]:
        return {
            "model": {
                "name": self.model_description.modelName,
                "description": self.model_description.description,
                "fmi_version": self.model_description.fmiVersion,
                "guid": self.model_description.guid,
                "generation_tool": self.model_description.generationTool,
                "generation_date": self.model_description.generationDateAndTime,
            },
            "inputs": self._inputs(),
            "outputs": self._outputs(),
            "parameters": self._parameters(),
        }

    def initialize(
        self,
        *,
        start_values: Optional[dict[str, FMUInputType]] = None,
        start_time: Union[int, float] = 0,
        step_size: Union[int, float] = 1,
    ) -> None:
        """Initialize the FMU.

        In this state, equations are active to determine all outputs (and optionally other variables
        exposed by the exporting tool). The variables that can be retrieved by fmi2GetXXX calls are
        (1) defined in the xml file under <ModelStructure><InitialUnknowns>, and
        (2) variables with causality = ′′output′′. Variables with initial = ′′exact′′ and
        variability ≠ "constant", as well as variables with causality = ′′input′′ can be set.
        """
        if not self._instantiated:
            self._instantiate()

        self._current_time = start_time
        self._step_size = step_size

        # Initialize FMU
        self.fmu.setupExperiment(startTime=self._current_time)

        start_values = start_values or {}
        self.fmu.enterInitializationMode()
        fmpy.simulation.apply_start_values(
            self.fmu, self.model_description, start_values=start_values or {}
        )
        self.fmu.exitInitializationMode()
        self._initialized = True

    def close(self) -> None:
        if self._initialized:
            self.fmu.terminate()
            self.fmu.reset()
        self._initialized = False

    def _free_instance(self) -> None:
        if self._instantiated:
            self.fmu.freeInstance()
            shutil.rmtree(self._unzipdir, ignore_errors=True)

    def reset(
        self,
        *,
        start_values: Optional[dict[str, FMUInputType]] = None,
        start_time: Union[int, float] = 0,
        step_size: Union[int, float] = 1,
    ) -> None:
        self._check_is_initialized()
        self.close()
        self.initialize(start_values=start_values, start_time=start_time, step_size=step_size)

    def _check_is_initialized(self):
        if not self._initialized:
            raise RuntimeError("FMU is not initialized.")

    def step(
        self, *, input_values: Optional[dict[str, FMUInputType]] = None
    ) -> dict[str, FMUInputType]:
        self._check_is_initialized()
        self._set_inputs(input_values or {})
        self._do_step()
        outputs = self.read_outputs()
        return outputs

    def advance(
        self, until: int, *, input_values: Optional[dict[str, FMUInputType]] = None
    ) -> dict[str, FMUInputType]:
        self._check_is_initialized()
        if until < self._current_time:
            raise ValueError("Cannot advance time to a time in the past.")

        self._set_inputs(input_values or {})

        while self._current_time < until:
            self._do_step()

        outputs = self.read_outputs()
        return outputs

    def _do_step(self):
        self.fmu.doStep(
            currentCommunicationPoint=self._current_time,
            communicationStepSize=self._step_size,
        )
        self._current_time += self._step_size

    def _set_inputs(self, input_values: dict[str, FMUInputType]) -> None:
        for input_name, input_value in input_values.items():
            _input = self._input_variables[input_name]
            self._write_variable(_input, input_value)

    def _write_variable(self, variable: ScalarVariable, value: FMUInputType) -> None:
        variable_type = variable.type
        variable_reference = variable.valueReference

        if variable_type == "Real":
            self.fmu.setReal([variable_reference], [float(value)])
        elif variable_type == "Integer":
            self.fmu.setInteger([variable_reference], [int(value)])
        elif variable_type == "Boolean":
            self.fmu.setBoolean([variable_reference], [bool(value)])
        elif variable_type == "String":
            self.fmu.setString([variable_reference], [str(value)])
        elif variable_type == "Enumeration":
            self.fmu.setInteger([variable_reference], [int(value)])
        else:
            raise ValueError(
                f"Unknown variable type '{variable_type}' for variable '{variable.name}'."
            )

    def read_outputs(self) -> dict[str, FMUInputType]:
        assert self.fmu is not None

        outputs = {
            "current_time": self._current_time,
        }
        if self._output_refs["Real"]:
            real_outputs = self.fmu.getReal(self._output_refs["Real"])
            outputs.update(dict(zip(self._output_names["Real"], real_outputs, strict=True)))

        if self._output_refs["Integer"]:
            integer_outputs = self.fmu.getInteger(self._output_refs["Integer"])
            outputs.update(dict(zip(self._output_names["Integer"], integer_outputs, strict=True)))

        if self._output_refs["Boolean"]:
            boolean_outputs = self.fmu.getBoolean(self._output_refs["Boolean"])
            outputs.update(
                {
                    output_name: bool(output_val)
                    for output_name, output_val in zip(
                        self._output_names["Boolean"], boolean_outputs, strict=True
                    )
                }
            )

        if self._output_refs["String"]:
            string_outputs = self.fmu.getString(self._output_refs["String"])
            outputs.update(
                {
                    output_name: str(output_val)
                    for output_name, output_val in zip(
                        self._output_names["String"], string_outputs, strict=True
                    )
                }
            )

        return outputs

    def _read_variable(self, variable: ScalarVariable) -> FMUInputType:
        variable_type = variable.type
        variable_reference = variable.valueReference

        if variable_type == "Real":
            return self.fmu.getReal([variable_reference])[0]
        elif variable_type == "Integer":
            return self.fmu.getInteger([variable_reference])[0]
        elif variable_type == "Boolean":
            return self.fmu.getBoolean([variable_reference])[0]
        elif variable_type == "String":
            return self.fmu.getString([variable_reference])[0]
        elif variable_type == "Enumeration":
            return self.fmu.getInteger([variable_reference])[0]
        raise ValueError(
            f"Unknown variable type '{variable_type}' for variable '{variable.name}'."
        )

    def change_parameters(self, parameters: dict[str, FMUInputType]) -> None:
        start_values = {
            variable.name: self._read_variable(variable)
            for variable in self.model_description.modelVariables
        }
        start_values.update(parameters)
        self.reset(
            start_values=start_values, start_time=self._current_time, step_size=self._step_size
        )

    def _inputs(self) -> dict[str, dict[str, Any]]:
        return {
            i.name: {
                "type": i.type,
                "description": i.description,
                "unit": i.unit,
                "min_value": i.min,
                "max_value": i.max,
                "start_value": i.start,
                "variability": i.variability,
            }
            for i in self._input_variables.values()
        }

    def _outputs(self) -> dict[str, dict[str, Any]]:
        return {
            o.name: {
                "type": o.type,
                "description": o.description,
                "unit": o.unit,
                "min_value": o.min,
                "max_value": o.max,
                "start_value": o.start,
                "variability": o.variability,
            }
            for o in self._output_variables.values()
        }

    def _parameters(self) -> dict[str, dict[str, Any]]:
        return {
            p.name: {
                "type": p.type,
                "description": p.description,
                "unit": p.unit,
                "min_value": p.min,
                "max_value": p.max,
                "start_value": p.start,
                "variability": p.variability,
            }
            for p in self._model_parameters.values()
        }



from datetime import datetime
from pathlib import Path
from typing import Any, Optional
from uuid import uuid4
from zoneinfo import ZoneInfo

from cosimtlk import FMIWrapper
from cosimtlk.models import FMUInputType

Record = dict[str, Any]


class SimulatorService:
    def __init__(self):
        self._db: dict[str, Record] = {}

    def close(self) -> None:
        keys = list(self._db.keys())
        for key in keys:
            del self._db[key]

    def create(
        self,
        path: Path,
        *,
        start_values: Optional[dict[str, FMUInputType]] = None,
        start_time: int = 0,
        step_size: int = 1,
    ) -> Record:
        if not path.exists():
            raise FileNotFoundError(path)

        wrapper = FMIWrapper(path)
        fmu = wrapper(
            start_values=start_values or {},
            start_time=start_time,
            step_size=step_size,
        )

        _id = str(uuid4())
        self._db[_id] = {
            "id": _id,
            "fmu": path.stem,
            "simulator": fmu,
            "created_at": datetime.now(tz=ZoneInfo("UTC")).isoformat(),
        }
        return self.get(_id)

    def list(self) -> list[Record]:
        return [self.get(_id) for _id, simulator_record in self._db.items()]

    def get(self, id: str) -> Record:
        return self._db[id]

    def get_simulator(self, id: str) -> FMIWrapper:
        return self._db[id]["simulator"]

    def delete(self, id: str) -> None:
        del self._db[id]


simulator_service = SimulatorService()



from typing import Any, Optional

import requests

from cosimtlk.app.schemas import SimulatorModel
from cosimtlk.models import FMUInputType


class SimulatorClient:
    def __init__(self, base_url: str):
        self.base_url = base_url
        self.session = requests.Session()

    def __repr__(self):
        return f"SimulatorClient({self.base_url})"

    @classmethod
    def from_parts(cls, host: str = "127.0.0.1", port: int = 8000, *, secure: bool = False):
        return cls(f"http{'s' if secure else ''}://{host}:{port}")

    @staticmethod
    def default_headers(**kwargs) -> dict[str, str]:
        return {
            "accept": "application/json",
            "Content-Type": "application/json",
            **kwargs,
        }

    def _get(self, path: str, **kwargs) -> dict:
        response = self.session.get(self.base_url + path, **kwargs)
        if not response.ok:
            response.raise_for_status()
        return response.json()

    def _post(self, path: str, *, body: dict, **kwargs) -> dict:
        response = self.session.post(
            self.base_url + path,
            headers=self.default_headers(**kwargs.pop("headers", {})),
            json=body,
            **kwargs,
        )
        if not response.ok:
            response.raise_for_status()
        return response.json()

    def _put(self, path: str, *, body: dict, **kwargs) -> dict:
        response = self.session.put(self.base_url + path, json=body, **kwargs)
        if not response.ok:
            response.raise_for_status()
        return response.json()

    def _delete(self, path: str, **kwargs) -> None:
        response = self.session.delete(self.base_url + path, **kwargs)
        if not response.ok:
            response.raise_for_status()
        return None

    def list_fmus(self) -> dict[str, Any]:
        return self._get("/fmus")

    def get_fmu_info(self, fmu: str) -> dict[str, Any]:
        return self._get(f"/fmus/{fmu}/info")

    def list_simulators(self) -> list[SimulatorModel]:
        return [SimulatorModel(**simulator) for simulator in self._get("/simulators")]

    def create_simulator(
        self,
        path: str,
        *,
        start_values: Optional[dict[str, FMUInputType]] = None,
        start_time: int = 0,
        step_size: int = 1,
    ) -> SimulatorModel:
        params = {"fmu": path}
        body = {
            "start_values": start_values or {},
            "start_time": start_time,
            "step_size": step_size,
        }
        response = self._post("/simulators/", params=params, body=body)
        return SimulatorModel(**response)

    def get_simulator(self, id: str):
        response = self._get(f"/simulators/{id}")
        return SimulatorModel(**response)

    def delete_simulator(self, id: str) -> None:
        return self._delete(f"/simulators/{id}")

    def get_outputs(self, id: str) -> dict[str, FMUInputType]:
        return self._get(f"/simulators/{id}/outputs")

    def step(
        self, id: str, *, input_values: Optional[dict[str, FMUInputType]] = None
    ) -> dict[str, FMUInputType]:
        body = input_values or {}
        return self._post(f"/simulators/{id}/step", body=body)

    def advance(
        self,
        id: str,
        until: int,
        *,
        input_values: Optional[dict[str, FMUInputType]] = None,
    ):
        params = {"until": until}
        body = input_values or {}
        return self._post(f"/simulators/{id}/advance", params=params, body=body)

    def change_parameters(
        self,
        id: str,
        *,
        parameters: dict[str, FMUInputType],
    ):
        body = parameters
        return self._put(f"/simulators/{id}/parameters", body=body)

    def reset(
        self,
        id: str,
        *,
        start_values: Optional[dict[str, FMUInputType]] = None,
        start_time: int = 0,
        step_size: int = 1,
    ):
        body = {
            "start_values": start_values or {},
            "start_time": start_time,
            "step_size": step_size,
        }
        return self._post(f"/simulators/{id}/reset", body=body)
